
//		SaveConsole.cpp - replaces the internal SAVECONSOLE command


#define UNICODE					1
#define _UNICODE				1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>

#include "PlugIn.h"
#include "TakeCmd.h"


#define VER_MAJOR				1
#define VER_MINOR				4
#define VER_BUILD				7
#define VER_PATCH				3

#define PLUGIN_NAME				L"SaveConsole"

#ifdef _WIN64
#define PLATFORM				L"x64"
#else
#define PLATFORM				L"x86"
#endif

#define MAX_ARG_LEN				8192
#define SHORT_BUF_LEN			1024
#define MAX_FILENAME_LEN		32768


#define ERRHANDLE				GetStdHandle( STD_ERROR_HANDLE )


#define MACRO_EXP( arg )		#arg
#define MACRO_STR( arg )		MACRO_EXP( arg )



HMODULE Kernel32Handle			= NULL;


typedef ULONGLONG ( WINAPI _GetTickCount64Process ) ( void );
_GetTickCount64Process *GetTickCount64Proc				= NULL;			//	in Windows Vista and later



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( ERRHANDLE, shortbuf );

	return 0;
}



int GetErrorText( int err, LPTSTR outString, unsigned int Flags = 0 )
{
	if ( outString == NULL )
		return -1;


	HANDLE WinINetHandle		= NULL;
	DWORD FM_Flags				= FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK;

	
	if ( ( err >= 12000 ) && ( err <= 12199 ) ) {
		WinINetHandle = GetModuleHandle( L"wininet.dll" );
		if ( WinINetHandle )
			FM_Flags |= FORMAT_MESSAGE_FROM_HMODULE;
	}


	if ( err == -1 )
		wcscpy_s( outString, SHORT_BUF_LEN, L"Unknown error" );
	else if ( !FormatMessage( FM_Flags, WinINetHandle, err, 0, outString, SHORT_BUF_LEN, NULL ) ) {
		if ( Flags & 1 )
			outString[0] = '\0';
		else
			wsprintf( outString, L"Error %d", err );
	}

	size_t l = wcslen( outString );

	if ( l )
		while ( iswspace( outString[l - 1] ) ) {
			l--;
			outString[l] = '\0';
		}

	return 0;
}



int GetLastErrorText( LPTSTR outString )
{
	return GetErrorText( GetLastError(), outString );
}



LPTSTR _FilenamePart( LPTSTR Filename, LPTSTR outFilename )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;


	BOOL FoundSlash			= FALSE;
	size_t EndOfPathPart	= -1;


	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) ) {
			EndOfPathPart = i;
			FoundSlash = TRUE;
		}
		else if ( ( Filename[i] == ':' ) && !FoundSlash )
			EndOfPathPart = i;
	}

	if ( EndOfPathPart == -1 ) {
		if ( outFilename ) {
			wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename );
			return outFilename;
		}
		else
			return Filename;
	}

	if ( outFilename ) {
		wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename + EndOfPathPart + 1 );
		return outFilename;
	}
	else
		return Filename + EndOfPathPart + 1;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion, LPCSTR SourceFile, LPCSTR SourceFileVer )
{
	wchar_t File[MAX_FILENAME_LEN]		= L"";
	wchar_t Ver[SHORT_BUF_LEN]			= L"";


	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion ) {
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

		if ( SourceFile && SourceFileVer ) {
			MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, SourceFile, -1, File, MAX_FILENAME_LEN );
			MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, SourceFileVer, -1, Ver, SHORT_BUF_LEN );
			Printf( L"\t\t\t\t%s  %s\r\n", _FilenamePart( File, NULL), Ver );
		}
	}

	return 0;
}



#include "ParseArgs.cpp"



BOOL IsColonOrEquals( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '=' ) );
}



int ParseInt( LPTSTR inString, unsigned int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;

	unsigned int value	= 0;
	wchar_t *sep1		= NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return -1;
		else if ( !EndChar && ( *sep1 != '\0' ) )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return -1;
		else if ( !EndChar && ( *sep1 != '\0' ) )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}

	return -1;
}



//	SAVECON_INTVARS, if defined, includes support for the
//	_SAVECONFILE and _SAVECONMODE internal variables :
//	
#define SAVECON_INTVARS

//	SAVECON_SHORT_HELP_TEXT, if defined, includes a much shorter
//	syntax help text for the SAVECONSOLE command, with only those
//	options supported by TCC's native SAVECONSOLE :
//	
//	#define SAVECON_SHORT_HELP_TEXT

//	SAVECON_RTF_SUPPORT, if defined, includes support for
//	generating .RTF files :

#define SAVECON_RTF_SUPPORT

#include "SaveConsoleCmd.cpp"



#ifdef SAVECON_INTVARS

const wchar_t PluginFeatures[] = L"saveconsole,_saveconfile,_saveconmode";

#else	//	SAVECON_INTVARS

const wchar_t PluginFeatures[] = L"saveconsole";

#endif	//	SAVECON_INTVARS



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/saveconsole.html";
	piInfo.pszDescription	= (LPTSTR) L"Replaces the internal SAVECONSOLE command.";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}


DLLExports BOOL WINAPI InitializePlugin( void )
{
	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	if ( wcscmp( shortbuf, L"00" ) == 0 )
		Printf( L"%s v%u.%u.%u loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );


	GetTickCount64Proc			= NULL;

	Kernel32Handle	= GetModuleHandle( L"Kernel32.dll" );
	if ( Kernel32Handle )
		GetTickCount64Proc		= (_GetTickCount64Process*)	GetProcAddress( Kernel32Handle, "GetTickCount64" );

	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{

#ifdef NEWHELP_VERSION
	CloseHelpWindow();
#endif

	SaveCon_Shutdown();
	return 0;
}



#ifdef SAVECON_INTVARS

const int NumHelpTexts							= 3;
const wchar_t HelpFeature[NumHelpTexts][13]		= { L"saveconsole", L"_saveconfile", L"_saveconmode" };
const wchar_t *HelpText[NumHelpTexts]			= { SaveConsoleHelpText, _SaveConFileHelpText, _SaveConModeHelpText };

#else	//	SAVECON_INTVARS

const int NumHelpTexts							= 1;
const wchar_t HelpFeature[NumHelpTexts][13]		= { L"saveconsole" };
const wchar_t *HelpText[NumHelpTexts]			= { SaveConsoleHelpText };

#endif	//	SAVECON_INTVARS



#ifndef	NEWHELP_VERSION

LPTSTR UsageText( const wchar_t *inString )
{
	if ( !inString )
		return NULL;


	wchar_t *ch = wcschr( (wchar_t*) inString, '\n' );


	if ( !ch )
		return NULL;

	while ( *ch == '\n' )
		ch++;

	return ch;
}

#endif	//	NEWHELP_VERSION




DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( ( inString[0] == '@' ) || ( inString[0] == '_' ) )
		return 0;

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			wcscpy_s( outString, SHORT_BUF_LEN, UsageText( HelpText[i] ) );
			return 1;
		}
	}

	return 0;
}


#ifndef NEWHELP_VERSION

//		Simpler MessageBox() help, useful for smaller plugins which don't merit a full HTML help system:

DLLExports int WINAPI Help( LPTSTR inString )
{
	wchar_t WindowTitle[SHORT_BUF_LEN]		= L"";


	wsprintf( WindowTitle, L"%s Plugin Help", PLUGIN_NAME );

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			MessageBox( GetForegroundWindow(), HelpText[i], WindowTitle, MB_OK );
			return 1;
		}
	}

	return 0;
}

#endif		//	NEWHELP_VERSION
