
//			SAVECONSOLE command, and related functions and structures:


#ifndef SAVECONSOLE_VERSION
#define SAVECONSOLE_VERSION				1.4.7.1


#ifndef MACRO_STR
#define MACRO_EXP( arg )				#arg
#define MACRO_STR( arg )				MACRO_EXP( arg )
#endif


#define SAVECON_FLAG_FULL_WIDTH			0x0004
#define SAVECON_FLAG_NO_BOM				0x0008
#define SAVECON_FLAG_TITLE				0x0020
#define SAVECON_FLAG_LINE_NUMBERS		0x0040
#define SAVECON_FLAG_NO_METADATA		0x0400

#define SAVECON_MODE_DEFAULT			0
#define SAVECON_MODE_HTML				1
#define SAVECON_MODE_RTF				2

#define SAVECON_HP_FLAG_DISABLE_TYPE	1
#define SAVECON_HP_FLAG_DISABLE_TANYA	2


#define SAVECON_DEFAULT_BG_COLOR		0x00ddddee
#define SAVECON_DEFAULT_COLOR			0x0007

#define SAVECON_FONT_SIZE_MIN			6
#define SAVECON_FONT_SIZE_MAX			48
#define SAVECON_FONT_SIZE_DEFAULT		10


#define N_HTML_ENTITIES					7


#ifndef MACRO_STR
#define MACRO_EXP( arg )				#arg
#define MACRO_STR( arg )				MACRO_EXP( arg )
#endif


#ifndef SPACE
#define SPACE							0x20
#endif	// SPACE

#ifndef BOM
#define BOM								0xfeff
#endif	//	BOM

#define SAVECON_BUFFER_SIZE				16384


#define SECURITY_WIN32
#include <Security.h>
#include <CommCtrl.h>

#include "BufferedOutput.cpp"
#include "w3ccolors.cpp"


struct _SaveConArgs {
	wchar_t			*CharBuffer;
	WORD			*AttrBuffer;
	unsigned int	Rows;
	unsigned int	Columns;
	WORD			DefaultColor;
	unsigned int	BackColor;
	DWORD			*Palette;
	unsigned int	FontSize;
	wchar_t			Eol[3];
	SYSTEMTIME		Now;
	LPCTSTR			StyleSheet;
	LPCTSTR			EmbedStyle;
};




#ifdef SAVECON_SHORT_HELP_TEXT

const wchar_t SaveConsoleHelpText[]		= L"Save the console buffer to a text or HTML file.\n\
\n\
SAVECONSOLE /H /T /W filename\r\n\
\n\
\t/H   save as HTML\n\
\t/T   include a header\n\
\t/W   save full width of every line\n\
\n\
CLIP: and directory aliases are supported.  If no filename  \n\
is given, you\u2019ll get a \u2018Save As\u2019 dialog.\n\
\n\
The default mode is to save as plain text, unless the\n\
filename ends in .HTML or .HTM.\n";

#elif defined SAVECON_RTF_SUPPORT

const wchar_t SaveConsoleHelpText[]		= L"Save the console buffer to a text, HTML, or RTF file.  \n\
\n\
SAVECONSOLE /D /H /L /O /Q /R /T /V /W filename  \r\n\
\n\
\t/D\tdo not save colors\n\
\t/H\tsave as HTML\n\
\t/L\tline numbers\n\
\t/O\topen output file when finished  \n\
\t/Q\tquietly\n\
\t/R\tsave as RTF\n\
\t/T\tinclude a header line\n\
\t/V\tvisible area only\n\
\t/W\tfull width (text only)\n\
\n\
CLIP: and directory aliases are supported.  If no filename  \n\
is given, you\u2019ll get a \u2018Save As\u2019 dialog.\n\
\n\
The default mode is to save as plain text, unless the\n\
filename ends in .HTML, .HTM, or .RTF.\n";

#else	//	SAVECON_SHORT_HELP_TEXT

const wchar_t SaveConsoleHelpText[]		= L"Save the console buffer to a text or HTML file.  \n\
\n\
SAVECONSOLE /D /H /L /O /Q /T /V /W filename  \r\n\
\n\
\t/D\tdo not save colors\n\
\t/H\tsave as HTML\n\
\t/L\tline numbers\n\
\t/O\topen output file when finished  \n\
\t/Q\tquietly\n\
\t/T\tinclude a header line\n\
\t/V\tvisible area only\n\
\t/W\tfull width (text only)\n\
\n\
CLIP: and directory aliases are supported.  If no filename  \n\
is given, you\u2019ll get a \u2018Save As\u2019 dialog.\n\
\n\
The default mode is to save as plain text, unless the\n\
filename ends in .HTML or .HTM.\n";

#endif	// SAVECON_SHORT_HELP_TEXT


const wchar_t SaveConsoleOptNHelpText[]	= L"SAVECONSOLE /N:  \u2014 disable features\n\
\n\
\tB\tdo not write a Byte Order Mark\n\
\tC\tdo not save colors\n\
\tD\tdisable the \u2018Save As\u2019 dialog\n\
\tH\tdisable the custom hook procedure for the dialog\n\
\tI\tdo not create an RTF info block\n\
\tT\tdisable the \u2018Save as type:\u2019 control in the dialog\n";


#ifdef SAVECON_INTVARS

//		Help text for the internal variables :

const wchar_t _SaveConFileHelpText[] = L"_SAVECONFILE\r\n\
\n\
Returns the filename if the last SAVECONSOLE  \n\
succeeded, or *ERROR* if it failed.\n";


const wchar_t _SaveConModeHelpText[] = L"_SAVECONMODE\r\n\
\n\
Returns the last SAVECONSOLE mode.\n";


//		Constants used for the _SAVECONFILE internal variable :

const wchar_t SaveCon_ClipName[]				= L"CLIP:";
const wchar_t SaveCon_ErrorString[]				= L"*ERROR*";


#ifdef SAVECON_REGISTRY_DEFAULTS

//		Names of registry values :

const wchar_t SaveCon_RegVal_NoBOM[]			= L"No BOM";
const wchar_t SaveCon_RegVal_Encoding[]			= L"Output Encoding";
const wchar_t SaveCon_RegVal_LineEnds[]			= L"Line Ends";
const wchar_t SaveCon_RegVal_NoDialog[]			= L"Disable SaveAs Dialog";

#endif	//	SAVECON_REGISTRY_DEFAULTS


//		Global variables used for the internal variables :

LPTSTR SC_LastFile				= NULL;
unsigned int SC_LastMode		= -1;

#endif	//	SAVECON_INTVARS


//		Global variables used to subclass the filename combo box in the GetSaveFileName dialog :

WNDPROC Tanya_OldWinProc		= NULL;
HWND Tanya_MessagesTo			= NULL;



const wchar_t SaveCon_Extensions[3][6]			= { L".txt", L".html", L".rtf" };


const wchar_t EntityChar[N_HTML_ENTITIES]		= { '&', '<', '>', 0x18, 0x19, 0x1a, 0x1b };

const wchar_t EntityName[N_HTML_ENTITIES][8]	= { L"&amp;", L"&lt;", L"&gt;", L"&uarr;", L"&darr;", L"&rarr;", L"&larr;" };


#ifdef SAVECON_RTF_SUPPORT

#define SAVECON_NUM_SMASH_CHARS			118

struct _SC_Smash {
	wchar_t ch;
	char ascii[4];
};

_SC_Smash SaveCon_SmashChars[SAVECON_NUM_SMASH_CHARS]	= { { 0xa0, " " }, { 0xa1, "!" }, { 0xa2, "c" }, { 0xa3, "L" },
	{ 0xa5, "Y" }, { 0xa6, "|" }, { 0xa7, "S" }, { 0xa9, "(C)" }, { 0xaa, "a" }, { 0xab, "<<" }, { 0xad, "-", },
	{ 0xae, "(R)" }, { 0xb0, "deg" }, { 0xb1, "" }, { 0xb2, "^2" }, { 0xb3, "^3" }, { 0xb4, "'" }, { 0xb5, "u" },
	{ 0xb6, "P" }, { 0xb7, "." }, { 0xb8, "," }, { 0xb9, "^1" }, { 0xba, "o" }, { 0xbb, ">>" }, { 0xbc, "1/4" },
	{ 0xbd, "1/2" }, { 0xbe, "3/4" }, { 0xbf, "?" }, { 0xc0, "A" }, { 0xc1, "A" }, { 0xc2, "A" }, { 0xc3, "A" }, 
	{ 0xc4, "A" }, { 0xc5, "A" }, { 0xc6, "AE" }, { 0xc7, "C" }, { 0xc8, "E" }, { 0xc9, "E" }, { 0xca, "E" },
	{ 0xcb, "E" }, { 0xcc, "I" }, { 0xcd, "I" }, { 0xce, "I" }, { 0xcf, "I" }, { 0xd0, "D" }, { 0xd1, "N" },
	{ 0xd2, "O" }, { 0xd3, "O" }, { 0xd4, "O" }, { 0xd5, "O" }, { 0xd6, "O" }, { 0xd7, "x" }, { 0xd8, "O" }, 
	{ 0xd9, "U" }, { 0xda, "U" }, { 0xdb, "U" }, { 0xdc, "U" }, { 0xdd, "Y" }, { 0xde, "Y" }, { 0xdf, "ss" },
	{ 0xe0, "a" }, { 0xe1, "a" }, { 0xe2, "a" }, { 0xe3, "a" }, { 0xe4, "a" }, { 0xe5, "a" }, { 0xe6, "ae" }, 
	{ 0xe7, "c" }, { 0xe8, "e" }, { 0xe9, "e" }, { 0xea, "e" }, { 0xeb, "e" }, { 0xec, "i" }, { 0xed, "i" }, 
	{ 0xee, "i" }, { 0xef, "i" }, { 0xf0, "d" }, { 0xf1, "n" }, { 0xf2, "o" }, { 0xf3, "o" }, { 0xf4, "o" }, 
	{ 0xf5, "o" }, { 0xf6, "o" }, { 0xf7, "/" }, { 0xf8, "o" }, { 0xf9, "u" }, { 0xfa, "u" }, { 0xfb, "o" }, 
	{ 0xfc, "u" }, { 0xfd, "y" }, { 0xfe, "y" }, { 0xff, "y" }, 
	{ 0x2000, " " }, { 0x2001, " " }, { 0x2002, " " }, { 0x2003, " " }, { 0x2004, " " }, { 0x2005, " " }, { 0x2006, " " }, 
	{ 0x2010, "-" }, { 0x2011, "-" }, { 0x2012, "-" }, { 0x2013, "-" }, { 0x2014, "--" }, { 0x2015, "--" }, 
	{ 0x2018, "'" }, { 0x2019, "'" }, { 0x201a, "'" }, { 0x201b, "'" }, { 0x201c, "\"" }, { 0x201d, "\"" }, 
	{ 0x201e, "\"" }, { 0x201f, "\"" }, { 0x2020, "+" }, { 0x2021, "+" }, { 0x2022, "*" }, { 0x2026, "..." }, 
	{ 0x20ac, "EUR" }, 
};

#endif	//	SAVECON_RTF_SUPPORT


const DWORD CssColors[16]		= { 0x000000, 0x000099, 0x009900, 0x009999, 0x990000, 0x990099, 0x996600, 0xcccccc,
									0x666666, 0x3333ff, 0x00ff00, 0x00ffff, 0xff0000, 0xff00ff, 0xffff00, 0xffffff };


#ifdef SAVECON_RTF_SUPPORT

const DWORD RtfColors[16]		= { 0x000000, 0x000080, 0x008000, 0x008080, 0x800000, 0x800080, 0x808000, 0xc0c0c0,
									0x808080, 0x0000ff, 0x00ff00, 0x00ffff, 0xff0000, 0xff00ff, 0xffff00, 0xffffff };

const int EgaToRtfColor[16]		= { 1, 9, 11, 10, 13, 12, 14, 16, 15, 2, 4, 3, 6, 5, 7, 8 };

const int RtfToEgaColor[16]		= { 0, 9, 11, 10, 13, 12, 14, 15, 1, 3, 2, 5, 4, 6, 8, 7 };

#endif	//	SAVECON_RTF_SUPPORT


unsigned __int64 GetTickCount2()
{
	if ( GetTickCount64Proc )
		return GetTickCount64Proc();
	else
		return GetTickCount();
}



int SaveCon_Shutdown()
{

#ifdef SAVECON_INTVARS

	free( SC_LastFile );
	SC_LastFile = NULL;

#endif	//	SAVECON_INTVARS

	return 0;
}



int SaveCon_Commafy( unsigned __int64 Value, LPTSTR outString )
{
	if ( !outString )
		return -1;

	wsprintf( outString, L"%%@comma[%I64u]", Value );
	ExpandVariables( outString, 1 );
	return 0;
}



LPTSTR SaveCon_ExtensionPart( LPTSTR Filename, LPTSTR outExt )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;


	LPTSTR Ext			= NULL;


	if ( outExt )
		outExt[0] = '\0';

	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) || ( Filename[i] == ':' ) )
			Ext = NULL;
		else if ( Filename[i] == '.' )
			Ext = Filename + i;
	}

	if ( !Ext )
		return NULL;

	if ( outExt ) {
		wcscpy_s( outExt, MAX_FILENAME_LEN, Ext );
		return outExt;
	}

	return Ext;
}



HWND SaveCon_GetTCWindow()
{
	wchar_t temp[SHORT_BUF_LEN]		= L"";

	HWND rv							= NULL;


	TakeCommandIPC( L"HWND", temp );

	if ( iswdigit( temp[0] ) )
		rv = (HWND) wcstoul( temp, NULL, 10 );

	return rv;
}



/*
		'Tanya' subclasses the filename combo-box control to catch EN_CHANGE notifications.
		When the text in the control is edited, Tanya posts a message back up to the
		GetSaveFileName() dialog, so it can check for changes to the file extension and
		handle them appropriately.
*/


LRESULT TanyaProc( HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam )
{
	if ( ( Msg == WM_COMMAND ) && ( HIWORD( wParam ) == EN_CHANGE ) )
		PostMessage( Tanya_MessagesTo, Msg, wParam, lParam );

	return Tanya_OldWinProc( hWnd, Msg, wParam, lParam );
}



UINT_PTR SaveCon_DialogHookProc$( HWND hWnd, UINT Msg, WPARAM Arg3, LPARAM Arg4 )
{
	static OPENFILENAME *OpenFileData	= NULL;
	static HWND DialogWindow			= NULL;
	static HWND FilenameComboBox		= NULL;
	static HWND FilenameComboBox2		= NULL;
	static HWND FilenameEditControl		= NULL;
	static HWND FileTypeDropDown		= NULL;
	static unsigned int SaveType		= 0;
 
	static UINT SelChangeMsg			= RegisterWindowMessageA( "commdlg_LBSelChangedNotify" );


	if ( Msg == WM_INITDIALOG ) {
		OpenFileData			= (OPENFILENAME*) Arg4;
		DialogWindow			= GetParent( hWnd );

		FilenameComboBox		= NULL;
		FilenameComboBox2		= NULL;
		FilenameEditControl		= NULL;
		FileTypeDropDown		= NULL;
		SaveType				= 0;

		FilenameComboBox = GetDlgItem( DialogWindow, cmb13 );
		FileTypeDropDown = GetDlgItem( DialogWindow, cmb1 );

		if ( FilenameComboBox ) {
			FilenameComboBox2	= (HWND) SendMessage( FilenameComboBox, CBEM_GETCOMBOCONTROL, 0, 0 );
			FilenameEditControl	= (HWND) SendMessage( FilenameComboBox, CBEM_GETEDITCONTROL, 0, 0 );
		}
		
		Tanya_OldWinProc		= NULL;
		Tanya_MessagesTo		= NULL;


		unsigned int Flags				= (unsigned int) OpenFileData->lCustData;
		SaveType						= OpenFileData->nFilterIndex - 1;


		if ( ( Flags & SAVECON_HP_FLAG_DISABLE_TYPE ) && FileTypeDropDown )
			EnableWindow( FileTypeDropDown, FALSE );
		else if ( FilenameComboBox2 && ( ( Flags & SAVECON_HP_FLAG_DISABLE_TANYA ) == 0 ) ) {
			Tanya_OldWinProc	= (WNDPROC) GetWindowLongPtr( FilenameComboBox2, GWLP_WNDPROC );

			if ( Tanya_OldWinProc ) {
				Tanya_MessagesTo	= hWnd;
				SetWindowLongPtr( FilenameComboBox2, GWLP_WNDPROC, (LONG_PTR) TanyaProc );
			}
		}
	}
	else if ( Msg == SelChangeMsg ) {

		if ( ( Arg3 == cmb1 ) && ( Arg4 > 0 ) && ( Arg4 < 4 ) && FilenameEditControl ) {

			wchar_t Filename[SHORT_BUF_LEN]			= L"";
			wchar_t *ch								= NULL;

			SaveType = (unsigned int) Arg4;

			GetWindowText( FilenameEditControl, Filename, SHORT_BUF_LEN );

			if ( Filename[0] ) {
				if ( ch = wcsrchr( Filename, '.' ) )
					*ch = '\0';

				wcscat_s( Filename, SHORT_BUF_LEN, SaveCon_Extensions[Arg4 - 1] );
				SetWindowText( FilenameEditControl, Filename );
			}
		}
	}
	else if ( ( Msg == WM_COMMAND ) && ( HIWORD( Arg3 ) == EN_CHANGE ) && ( Arg4 == (LPARAM) FilenameEditControl ) ) {
		if ( FileTypeDropDown && FilenameEditControl ) {

			wchar_t Filename[SHORT_BUF_LEN]		= L"";

			wchar_t *ch				= NULL;
			unsigned int NewType	= 0;

			GetWindowText( FilenameEditControl, Filename, SHORT_BUF_LEN );

			if ( Filename[0] ) {
				if ( ch = wcsrchr( Filename, '.' ) ) {

					if ( !_wcsicmp( ch, L".txt" ) )
						NewType = 1;
					else if ( !_wcsicmp( ch, L".html" ) )
						NewType = 2;
					else if ( !_wcsicmp( ch, L".htm" ) )
						NewType = 2;

#ifdef SAVECON_RTF_SUPPORT

					else if ( !_wcsicmp( ch, L".rtf" ) )
						NewType = 3;

#endif	//	SAVECON_RTF_SUPPORT

					if ( NewType && ( NewType != SaveType ) ) {
						SendMessage( FileTypeDropDown, CB_SETCURSEL, NewType, 0 );
						SaveType = NewType;
					}
				}
			}
		}
	}
	else if ( Msg == WM_NOTIFY ) {
		NMHDR *NotifyHeader = (NMHDR*) Arg4;

		if ( ( NotifyHeader->code == CDN_FILEOK ) && OpenFileData )
			OpenFileData->nFilterIndex = SaveType + 1;
	}

	return 0;
}



int SaveCon_GetFilename( LPTSTR Filename, size_t BufLen, BOOL UseHook, unsigned int *Mode, unsigned int Flags )
{
	if ( !Filename || ( BufLen < 16 ) || !Mode )
		return -1;


	wchar_t FileNameBuf[MAX_FILENAME_LEN]		= L"";
	wchar_t InitialDir[MAX_FILENAME_LEN]		= L"";

#ifdef SAVECON_RTF_SUPPORT

	const wchar_t Filters[]						= L"All Files\0*.*\0Text Files\0*.txt\0HTML Files\0*.html;*.htm\0RTF Files\0*.rtf\0\0";

#else	//	SAVECON_RTF_SUPPORT

	const wchar_t Filters[]						= L"All Files\0*.*\0Text Files\0*.txt\0HTML Files\0*.html;*.htm\0\0";

#endif	//	SAVECON_RTF_SUPPORT

	unsigned int InitFilterIndex				= *Mode + 2;
	HWND OwnerWindow							= SaveCon_GetTCWindow();

	OPENFILENAME OpenFileData;


	if ( Filename[0] )
		wcscpy_s( InitialDir, MAX_FILENAME_LEN, Filename );
	else
		GetCurrentDirectory( MAX_FILENAME_LEN, InitialDir );


	if ( !OwnerWindow )
		OwnerWindow = GetConsoleWindow();

	if ( !IsWindowVisible( OwnerWindow ) )
		OwnerWindow = GetForegroundWindow();


	memset( &OpenFileData, 0x00, sizeof( OPENFILENAME ) );

	OpenFileData.lStructSize		= sizeof( OPENFILENAME );
	OpenFileData.hwndOwner			= OwnerWindow;
	OpenFileData.lpstrFilter		= Filters;
	OpenFileData.nFilterIndex		= InitFilterIndex;
	OpenFileData.lpstrFile			= FileNameBuf;
	OpenFileData.nMaxFile			= MAX_FILENAME_LEN;
	OpenFileData.lpstrInitialDir	= InitialDir;
	OpenFileData.lCustData			= (LPARAM) Flags;


	if ( UseHook ) {
		OpenFileData.Flags				= OFN_NOCHANGEDIR | OFN_ENABLEHOOK | OFN_EXPLORER | OFN_ENABLESIZING;
		OpenFileData.lpfnHook			= (LPOFNHOOKPROC) SaveCon_DialogHookProc$;
	}
	else {
		OpenFileData.Flags				= OFN_NOCHANGEDIR;
	}


	if ( !GetSaveFileName( &OpenFileData ) ) {
		Filename[0] = '\0';
		return 3;
	}

	if ( OpenFileData.lpstrFile[0] == '\0' ) {
		Filename[0] = '\0';
		return 1;
	}


	wcscpy_s( Filename, BufLen, OpenFileData.lpstrFile );

	if ( ( Flags & SAVECON_HP_FLAG_DISABLE_TYPE ) == 0 )
		if ( OpenFileData.nFilterIndex >= 2 )
			*Mode = OpenFileData.nFilterIndex - 2;

	return 0;
}



#ifdef SAVECON_REGISTRY_DEFAULTS

int SaveCon_GetDefaults( HKEY RootKey, unsigned int *Flags, int *OutputEncoding, wchar_t* LineEnds, BOOL *NoDialog )
{
	if ( !RootKey || !Flags || !OutputEncoding || !LineEnds || !NoDialog )
		return -1;

	wchar_t RegKeyName[SHORT_BUF_LEN]			= L"";

	HKEY KeyHandle			= NULL;
	DWORD valtype			= 0;
	DWORD valsize			= 0;
	DWORD value				= 0;

	int rv					= 0;


	wcscpy_s( RegKeyName, SHORT_BUF_LEN, L"SOFTWARE\\JPPlugins\\" );
	wcscat_s( RegKeyName, SHORT_BUF_LEN, PLUGIN_NAME );

	if ( RegOpenKeyEx( RootKey, RegKeyName, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {

		valtype	= 0;
		valsize	= sizeof( DWORD );
		value	= 0;

		if ( RegQueryValueEx( KeyHandle, SaveCon_RegVal_NoBOM, NULL, &valtype, (BYTE*) &value, &valsize ) )
			value = -1;
		else if ( ( valtype != REG_DWORD ) || ( valsize != sizeof( DWORD ) ) )
			value	= -1;
		else if ( value ) {
			rv++;
			*Flags |= SAVECON_FLAG_NO_BOM;
		}
		else {
			rv++;
			*Flags &= ~SAVECON_FLAG_NO_BOM;
		}


		valtype	= 0;
		valsize	= sizeof( DWORD );
		value	= 0;

		if ( RegQueryValueEx( KeyHandle, SaveCon_RegVal_Encoding, NULL, &valtype, (BYTE*) &value, &valsize ) )
			value = -1;
		else if ( ( valtype != REG_DWORD ) || ( valsize != sizeof( DWORD ) ) )
			value = -1;
		else if ( value == 1 )
			*OutputEncoding = BUFOUT_FILE_UTF8, rv++;
		else if ( value == 2 )
			*OutputEncoding = BUFOUT_FILE_UTF16, rv++;


		valtype	= 0;
		valsize	= sizeof( DWORD );
		value	= 0;

		if ( RegQueryValueEx( KeyHandle, SaveCon_RegVal_LineEnds, NULL, &valtype, (BYTE*) &value, &valsize ) )
			value = -1;
		else if ( ( valtype != REG_DWORD ) || ( valsize != sizeof( DWORD ) ) )
			value = -1;
		else {
			switch ( value ) {

			case 1 :
				LineEnds[0] = '\n', LineEnds[1] = '\0', rv++;
				break;

			case 2 :
				LineEnds[0] = '\r', LineEnds[1] = '\n', LineEnds[2] = '\0', rv++;
				break;

			case 3 :
				LineEnds[0] = '\r', LineEnds[1] = '\0', rv++;
				break;
			}
		}



		valtype	= 0;
		valsize	= sizeof( DWORD );
		value	= 0;

		if ( RegQueryValueEx( KeyHandle, SaveCon_RegVal_NoDialog, NULL, &valtype, (BYTE*) &value, &valsize ) )
			value	= -1;
		else if ( ( valtype != REG_DWORD ) || ( valsize != sizeof( DWORD ) ) )
			value	= -1;
		else {
			rv++;
			*NoDialog = ( value > 0 );
		}

		RegCloseKey( KeyHandle );
	}

	return rv;
}



HKEY SaveCon_GetDefaults( unsigned int *Flags, int *OutputEncoding, wchar_t *LineEnds, BOOL *NoDialog )
{
	if ( SaveCon_GetDefaults( HKEY_CURRENT_USER, Flags, OutputEncoding, LineEnds, NoDialog ) )
		return HKEY_CURRENT_USER;

	if ( SaveCon_GetDefaults( HKEY_LOCAL_MACHINE, Flags, OutputEncoding, LineEnds, NoDialog ) )
		return HKEY_LOCAL_MACHINE;

	return NULL;
}

#endif	//	SAVECON_REGISTRY_DEFAULTS



BOOL IsConsoleHandle( HANDLE Handle )
{
	DWORD Test			= -1;


	if ( !GetConsoleMode( Handle, &Test ) )
		return FALSE;

	if ( Test == (DWORD) -1 )
		return FALSE;

	return TRUE;
}



BOOL SaveCon_IsAFile( LPCTSTR Filename )
{
	if ( !Filename )
		return FALSE;
	else if ( Filename[0] == '\0' )
		return FALSE;


	DWORD attr		= GetFileAttributes( Filename );


	if ( attr == INVALID_FILE_ATTRIBUTES )
		return FALSE;

	if ( attr & ( FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_DEVICE ) )
		return FALSE;

	return TRUE;
}



LPCTSTR SaveCon_EolType( LPCTSTR Eol )
{
	static const wchar_t eol[5][6]		= { L"LF", L"CR/LF", L"CR", L"NEL", L"?" };


	if ( !wcscmp( Eol, L"\n" ) )
		return eol[0];
	else if ( !wcscmp( Eol, L"\r\n" ) )
		return eol[1];
	else if ( !wcscmp( Eol, L"\r" ) )
		return eol[2];
	else if ( !wcscmp( Eol, L"\x85" ) )
		return eol[3];
	else
		return eol[4];
}



wchar_t SaveCon_LastChar( LPCTSTR String )
{
	if ( !String )
		return '\0';

	size_t l		= wcslen( String );

	if ( !l )
		return '\0';
	else
		return String[l - 1];
}



int SaveCon_LaunchFile( LPTSTR Filename )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	SHELLEXECUTEINFO ShellExecInfo;
	int rv			= 0;


	memset( &ShellExecInfo, 0x00, sizeof( SHELLEXECUTEINFO ) );

	ShellExecInfo.cbSize		= sizeof( SHELLEXECUTEINFO );
	ShellExecInfo.lpFile		= Filename;
	ShellExecInfo.nShow			= SW_SHOW;

	if ( !ShellExecuteEx( &ShellExecInfo ) )
		rv = GetLastError();

	return rv;
}



int SaveCon_CopyToClipboard( HANDLE Handle, unsigned int Format )
{
	if ( !Handle ) {
		SetLastError( ERROR_INVALID_HANDLE );
		return 0;
	}


	LARGE_INTEGER FileSize;
	HANDLE MemHandle				= NULL;
	BYTE *Buffer					= NULL;
	const unsigned int ReadSize		= 8192;
	size_t ReadAddress				= 0;
	size_t BytesRemaining			= 0;
	DWORD BytesToRead				= 0;
	DWORD BytesRead					= 0;
	int ReadErrors					= 0;

	int rv							= 0;



	memset( &FileSize, 0x00, sizeof( LARGE_INTEGER ) );

	if ( !GetFileSizeEx( Handle, &FileSize ) )
		return 0;

	if ( FileSize.QuadPart == 0 ) {
		SetLastError( ERROR_NO_DATA );
		return 0;
	}

	SetFilePointer( Handle, 0, NULL, FILE_BEGIN );


	MemHandle = GlobalAlloc( GHND, (size_t) FileSize.QuadPart );

	if ( MemHandle ) {
		Buffer = (BYTE*) GlobalLock( MemHandle );

		if ( Buffer ) {
			BytesRemaining = (size_t) FileSize.QuadPart;

			while ( BytesRemaining ) {

				if ( BytesRemaining > (size_t) ReadSize )
					BytesToRead = ReadSize;
				else
					BytesToRead = (DWORD) BytesRemaining;

				BytesRead = 0;

				if ( !ReadFile( Handle, Buffer + ReadAddress, BytesToRead, &BytesRead, NULL ) )
					ReadErrors++;
				else if ( !BytesRead )
					ReadErrors++;
				else {
					ReadAddress += BytesRead;
					BytesRemaining -= BytesRead;
				}
			};

			GlobalUnlock( MemHandle );

			if ( !ReadErrors ) {
				if ( OpenClipboard( GetConsoleWindow() ) ) {
					if ( EmptyClipboard() )
						if ( SetClipboardData( Format, MemHandle ) )
							rv = 1;

					CloseClipboard();
				}
			}
		}
	}

	return rv;
}



int SaveCon_CopyToClipboard( LPTSTR Filename, unsigned int Format )
{
	if ( !Filename ) {
		SetLastError( ERROR_INVALID_PARAMETER );
		return 0;
	}


	HANDLE FileHandle			= NULL;
	int rv						= 0;


	FileHandle = CreateFile( Filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL );
	if ( FileHandle == INVALID_HANDLE_VALUE )
		return 0;

	rv = SaveCon_CopyToClipboard( FileHandle, Format );
	CloseHandle( FileHandle );
	return rv;
}



LPTSTR SaveCon_EntityName( unsigned int ch )
{
	for ( unsigned int i = 0; i < N_HTML_ENTITIES; i++ )
		if ( ch == EntityChar[i] )
			return (LPTSTR) EntityName[i];

	return NULL;
}



int SaveCon_Entitize( LPTSTR inString, LPTSTR outString, size_t BufLen )
{
	if ( !inString || !outString )
		return -1;


	wchar_t temp[MAX_ARG_LEN]		= L"";
	size_t j						= 0;
	LPTSTR Entity					= NULL;


	for ( size_t i = 0; inString[i]; i++ ) {
		Entity = SaveCon_EntityName( inString[i] );

		if ( Entity ) {
			wcscat_s( temp, MAX_ARG_LEN, Entity );
			j = wcslen( temp );
		}
		else {
			if ( j < ( MAX_ARG_LEN - 1 ) ) {
				temp[j++] = inString[i];
				temp[j] = '\0';
			}
		}
	}

	if ( wcslen( temp ) >= BufLen )
		return ERROR_NOT_ENOUGH_MEMORY;

	wcscpy_s( outString, BufLen, temp );
	return 0;
}



DWORD SaveCon_Rgb2Bgr( DWORD Color )
{
	return ( ( Color & 0x0000ff ) << 16 ) | ( Color & 0x00ff00 ) | ( ( Color & 0xff0000 ) >> 16 );
}



int SaveCon_GetConsolePalette( HANDLE ConsoleHandle, DWORD *ConsolePalette )
{
	if ( !ConsoleHandle || !ConsolePalette )
		return 0;


	typedef BOOL ( WINAPI _GetCSBIExProc ) ( HANDLE ConsoleHandle, PCONSOLE_SCREEN_BUFFER_INFOEX BufferInfoEx );
	_GetCSBIExProc *GetConsoleBufferInfoExProcess			= NULL;			//  in Windows Vista and later

	CONSOLE_SCREEN_BUFFER_INFOEX BufferInfoEx;



	GetConsoleBufferInfoExProcess = (_GetCSBIExProc*) GetProcAddress( Kernel32Handle, "GetConsoleScreenBufferInfoEx" );
	if ( !GetConsoleBufferInfoExProcess )
		return 0;

	memset( &BufferInfoEx, 0x00, sizeof( CONSOLE_SCREEN_BUFFER_INFOEX ) );
	BufferInfoEx.cbSize = sizeof( CONSOLE_SCREEN_BUFFER_INFOEX );

	if ( !GetConsoleBufferInfoExProcess( ConsoleHandle, &BufferInfoEx ) )
		return 0;


	for ( unsigned int i = 0; i < 16; i++ )
		ConsolePalette[i]	= SaveCon_Rgb2Bgr( BufferInfoEx.ColorTable[i] );

	return 16;
}



LPCTSTR SaveCon_CompactRgb( DWORD Color )
{
	const int NumBuffers			= 8;

	static wchar_t Buffers[NumBuffers][8];
	static int k					= 0;

	LPTSTR ThisBuffer				= NULL;
	DWORD color3					= 0;



	ThisBuffer = Buffers[k];


	if ( ( ( Color & 0xfff0f0f0 ) >> 4 ) == ( Color & 0x000f0f0f ) ) {
		color3 = ( ( Color & 0x000f0000 ) >> 8 ) | ( ( Color & 0x00000f00 ) >> 4 ) | ( Color & 0x0000000f );
		wsprintf( ThisBuffer, L"#%03x", color3 );
	}
	else
		wsprintf( ThisBuffer, L"#%06x", Color );

	k = ( k + 1 ) % NumBuffers;
	return ThisBuffer;
}



int SaveCon_CreateStyleSheet( BufOut_Buffer *OutBuffer, unsigned int Flags, DWORD *ConsolePalette, _SaveConArgs *Args )
{
	wchar_t temp[SHORT_BUF_LEN]		= L"";

	unsigned int ColorFound[256];
	unsigned int fg					= Args->DefaultColor & 0x000f;
	unsigned int bg					= ( Args->DefaultColor & 0x00f0 ) >> 4;
	wchar_t est[2]					= L";";

	int rv							= 0;


	memset( ColorFound, 0x00, 256 * sizeof( unsigned int ) );


	if ( Args->AttrBuffer )
		for ( unsigned int i = 0; i < ( Args->Rows * Args->Columns ); i++ )
			ColorFound[ Args->AttrBuffer[i] & 0x00ff ] += 1;


	if ( Args->EmbedStyle )
		if ( SaveCon_LastChar( Args->EmbedStyle ) == ';' )
			est[0] = '\0';


	if ( rv = BufferedOutput( OutBuffer, L"<style>" ) ) return rv;
	if ( rv = BufferedOutput( OutBuffer, Args->Eol ) ) return rv;

	wsprintf( temp, L"\tBODY\t{ background-color: %s; }%s", SaveCon_CompactRgb( Args->BackColor ), Args->Eol );
	if ( rv = BufferedOutput( OutBuffer, temp ) ) return rv;

	if ( Args->EmbedStyle && Args->FontSize )
		wsprintf( temp, L"\tPRE \t{ font-size: %upt; %s%s display: inline-block; color: %s; background-color: %s; ", Args->FontSize, Args->EmbedStyle, est, SaveCon_CompactRgb( ConsolePalette[fg] ), SaveCon_CompactRgb( ConsolePalette[bg] ) );
	else if ( Args->EmbedStyle )
		wsprintf( temp, L"\tPRE \t{ %s%s display: inline-block; color: %s; background-color: %s; ", Args->EmbedStyle, est, SaveCon_CompactRgb( ConsolePalette[fg] ), SaveCon_CompactRgb( ConsolePalette[bg] ) );
	else
		wsprintf( temp, L"\tPRE \t{ display: inline-block; color: %s; background-color: %s; border: 4px inset %s; padding: 6px; margin-left: 5px; ", SaveCon_CompactRgb( ConsolePalette[fg] ), SaveCon_CompactRgb( ConsolePalette[bg] ), SaveCon_CompactRgb( ConsolePalette[fg] ) );

	if ( rv = BufferedOutput( OutBuffer, temp ) ) return rv;

	if ( Args->FontSize && !Args->EmbedStyle ) {
		wsprintf( temp, L"font-size: %upt; ", Args->FontSize );
		if ( rv = BufferedOutput( OutBuffer, temp ) ) return rv;
	}

	wsprintf( temp, L"}%s", Args->Eol );
	if ( rv = BufferedOutput( OutBuffer, temp ) ) return rv;

	if ( Flags & SAVECON_FLAG_TITLE ) {
		if ( rv = BufferedOutput( OutBuffer, L"\tH1  \t{ font-size: 1.2em; margin-left: 5px; }" ) ) return rv;
		if ( rv = BufferedOutput( OutBuffer, Args->Eol ) ) return rv;
	}


	if ( Args->AttrBuffer ) {
		for ( unsigned int i = 0; i < 256; i++ ) {
			if ( ColorFound[i] && ( i != Args->DefaultColor ) ) {

				wsprintf( temp, L"\t.X%02x\t{ color: %s; background-color: %s; }%s", i, SaveCon_CompactRgb( ConsolePalette[i & 0x0f] ), SaveCon_CompactRgb( ConsolePalette[( i & 0xff ) >> 4] ), Args->Eol );
				if ( rv = BufferedOutput( OutBuffer, temp ) ) return rv;
			}
		}
	}

	if ( rv = BufferedOutput( OutBuffer, L"</style>" ) ) return rv;
	if ( rv = BufferedOutput( OutBuffer, Args->Eol ) ) return rv;

	return 0;
}


int SaveCon_HtmlHeader( BufOut_Buffer *OutBuffer, LPTSTR Title, unsigned int Flags, DWORD *ConsolePalette, _SaveConArgs *Args )
{
	wchar_t temp[MAX_ARG_LEN]			= L"";
	wchar_t SourceVer[SHORT_BUF_LEN]	= L"";

	int rv				= 0;


	if ( Title )
		if ( Title[0] == '\0' )
			Title = NULL;


	if ( rv = BufferedOutput( OutBuffer, L"<!DOCTYPE html>" ) ) return rv;
	if ( rv = BufferedOutput( OutBuffer, Args->Eol ) ) return rv;

	wsprintf( temp, L"<html lang=\"zxx\">%s%s<head>%s", Args->Eol, Args->Eol, Args->Eol );

	if ( rv = BufferedOutput( OutBuffer, temp ) ) return rv;

	if ( OutBuffer->enc == BUFOUT_FILE_UTF8 )
		rv = BufferedOutput( OutBuffer, L"<meta charset=\"UTF-8\">" );
	else
		rv = BufferedOutput( OutBuffer, L"<meta charset=\"UTF-16LE\">" );

	if ( rv )
		return rv;

	if ( rv = BufferedOutput( OutBuffer, Args->Eol ) ) return rv;


	MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, MACRO_STR( SAVECONSOLE_VERSION ), -1, SourceVer, SHORT_BUF_LEN );
	wsprintf( temp, L"<!-- SaveConsole v%s  run %04u-%02u-%02u %02u:%02u:%02u -->%s", SourceVer, Args->Now.wYear, Args->Now.wMonth, Args->Now.wDay, Args->Now.wHour, Args->Now.wMinute, Args->Now.wSecond, Args->Eol );
	if ( rv = BufferedOutput( OutBuffer, temp ) )
		return rv;


	if ( Title ) {
		if ( rv = BufferedOutput( OutBuffer, L"<title>" ) ) return rv;
		if ( rv = BufferedOutput( OutBuffer, Title ) ) return rv;
		if ( rv = BufferedOutput( OutBuffer, L"</title>" ) ) return rv;
		if ( rv = BufferedOutput( OutBuffer, Args->Eol ) ) return rv;
	}

	if ( rv = SaveCon_CreateStyleSheet( OutBuffer, Flags, ConsolePalette, Args ) ) return rv;

	if ( Args->StyleSheet ) {
		if ( Args->StyleSheet[0] ) {
			wsprintf( temp, L"<link rel=\"stylesheet\" href=\"%s\">%s", Args->StyleSheet, Args->Eol );
			if ( rv = BufferedOutput( OutBuffer, temp ) ) return rv;
		}
	}

	if ( rv = BufferedOutput( OutBuffer, L"</head>" ) ) return rv;
	if ( rv = BufferedOutput( OutBuffer, Args->Eol ) ) return rv;
	if ( rv = BufferedOutput( OutBuffer, Args->Eol ) ) return rv;

	return 0;
}


int SaveCon_HtmlStartBody( BufOut_Buffer *OutBuffer, LPTSTR Title, LPCTSTR Eol )
{
	wchar_t temp[SHORT_BUF_LEN]			= L"";

	int rv								= 0;


	if ( rv = BufferedOutput( OutBuffer, L"<body>" ) ) return rv;
	if ( rv = BufferedOutput( OutBuffer, Eol ) ) return rv;

	if ( Title ) {
		if ( rv = BufferedOutput( OutBuffer, L"<h1>" ) ) return rv;
		if ( rv = BufferedOutput( OutBuffer, Title ) ) return rv;
		if ( rv = BufferedOutput( OutBuffer, L"</h1>" ) ) return rv;
		if ( rv = BufferedOutput( OutBuffer, Eol ) ) return rv;
	}

	if ( rv = BufferedOutput( OutBuffer, L"<pre>" ) ) return rv;

	return 0;
}


int SaveCon_HtmlEndBody( BufOut_Buffer *OutBuffer, LPCTSTR Eol )
{
	wchar_t temp[SHORT_BUF_LEN]		= L"";

	int rv							= 0;


	wsprintf( temp, L"</pre>%s</body>%s</html>%s", Eol, Eol, Eol );
	if ( rv = BufferedOutput( OutBuffer, temp ) ) return rv;

	return 0;
}



int SaveCon_MakeTitle( LPTSTR Buffer, SYSTEMTIME *Now )
{
	wchar_t Title[SHORT_BUF_LEN]	= L"";
	wchar_t datestr[SHORT_BUF_LEN]	= L"";
	wchar_t timestr[SHORT_BUF_LEN]	= L"";


	GetConsoleTitle( Title, SHORT_BUF_LEN );


	GetDateFormat( LOCALE_USER_DEFAULT, DATE_LONGDATE, Now, NULL, datestr, SHORT_BUF_LEN );
	GetTimeFormat( LOCALE_USER_DEFAULT, 0, Now, NULL, timestr, SHORT_BUF_LEN );

	
	wcscpy_s( Buffer, MAX_ARG_LEN, Title );
	wcscat_s( Buffer, MAX_ARG_LEN, L" \u2014 " );
	wcscat_s( Buffer, MAX_ARG_LEN, datestr );
	wcscat_s( Buffer, MAX_ARG_LEN, L" " );
	wcscat_s( Buffer, MAX_ARG_LEN, timestr );
	return 0;
}



int SaveCon_MakeFilename( LPTSTR Filename )
{
	wchar_t shortbuf[SHORT_BUF_LEN]			= L"";
	SYSTEMTIME Now;
	size_t len				= 0;
	static unsigned int k	= 0;



	GetTempPath( MAX_FILENAME_LEN, Filename );
	GetLocalTime( &Now );

	len		= wcslen( Filename );


	//	If the buffer contains a directory name and
	//	doesn't end with a backslash, add a backslash :

	if ( len )
		if ( Filename[len - 1] != '\\' ) {
			wcscat_s( Filename, MAX_FILENAME_LEN, L"\\" );
			len++;
		}


	//	First try a filename based on the date and time :

	wsprintf( shortbuf, L"savecon-%04d%02d%02d-%02d%02d%02d.tmp", Now.wYear, Now.wMonth, Now.wDay, Now.wHour, Now.wMinute, Now.wSecond );
	wcscat_s( Filename, MAX_FILENAME_LEN, shortbuf );

	if ( !SaveCon_IsAFile( Filename ) )
		return 0;


	//	If a file by that name exists, try numeric suffixes
	//	until an unused filename is found :

	do {
		Filename[len] = '\0';
		wsprintf( shortbuf, L"savecon-%05x.tmp", k++ );
		wcscat_s( Filename, MAX_FILENAME_LEN, shortbuf );
	}
	while ( SaveCon_IsAFile( Filename ) );

	return 0;
}



int SaveCon_Html( HANDLE ConsoleHandle, BufOut_Buffer *OutBuffer, unsigned int Flags, _SaveConArgs *Args )
{
	wchar_t Title[MAX_ARG_LEN]			= L"";
	wchar_t temp[SHORT_BUF_LEN]			= L"";

	BOOL SaveColors						= Args->AttrBuffer ? TRUE : FALSE;
	BOOL InColorSpan					= FALSE;

	unsigned int NumColumns				= Args->Columns;

	unsigned int ch						= '\0';
	LPTSTR Entity						= NULL;
	WORD CellColor						= 0;
	WORD LastColor						= -1;
	WORD sbgtest						= Args->DefaultColor & 0x00f0;

	int rv								= 0;



	SaveCon_MakeTitle( Title, &Args->Now );
	SaveCon_Entitize( Title, Title, MAX_ARG_LEN );


	if ( ( Flags & SAVECON_FLAG_NO_BOM ) == 0 )
		BufferedOutput( OutBuffer, BOM );

	if ( rv = SaveCon_HtmlHeader( OutBuffer, Title, Flags, Args->Palette, Args ) ) return rv;

	if ( rv = SaveCon_HtmlStartBody( OutBuffer, ( Flags & SAVECON_FLAG_TITLE ) ? Title : NULL, Args->Eol ) ) return rv;


	for ( unsigned int i = 0; i < Args->Rows; i++ ) {
		NumColumns = Args->Columns;

		if ( i ) {
			if ( SaveColors ) {
				while ( NumColumns && ( Args->CharBuffer[( i * Args->Columns ) + NumColumns - 1] == SPACE ) && ( ( Args->AttrBuffer[( i * Args->Columns ) + NumColumns -1] & 0x00f0 ) == sbgtest ) )
					NumColumns--;
			}
			else {
				while ( NumColumns && ( Args->CharBuffer[( i * Args->Columns ) + NumColumns - 1] == SPACE ) )
					NumColumns--;
			}
		}

		if ( Flags & SAVECON_FLAG_LINE_NUMBERS ) {
			if ( SaveColors && ( LastColor != Args->DefaultColor ) ) {
				if ( InColorSpan ) {
					if ( rv = BufferedOutput( OutBuffer, L"</span>" ) ) return rv;
					InColorSpan = FALSE;
				}

				LastColor = Args->DefaultColor;
			}

			wsprintf( temp, L"%4u:  ", i + 1 );
			if ( rv = BufferedOutput( OutBuffer, temp ) ) return rv;
		}

		for ( unsigned int j = 0; j < NumColumns; j++ ) {
			ch = Args->CharBuffer[( i * Args->Columns ) + j];
			Entity = SaveCon_EntityName( ch );
			
			//	Make sure we're not going to split a surrogate pair,
			//	as that would bollix UTF-16 to UTF-8 conversion :

			if ( ( ( ch & 0xfc00 ) == 0xd800 ) && ( j < ( NumColumns - 1 ) ) )
				if ( ( Args->CharBuffer[( i * Args->Columns ) + j + 1] & 0xfc00 ) == 0xdc00 )
					if ( OutBuffer->ptr > ( OutBuffer->size - 6 ) )
						if ( rv = FlushOutBuffer( OutBuffer ) ) return rv;

			if ( SaveColors ) {
				CellColor  = Args->AttrBuffer[( i * Args->Columns ) + j] & 0xff;

				if ( CellColor != LastColor ) {
					if ( InColorSpan ) {
						if ( rv = BufferedOutput( OutBuffer, L"</span>" ) ) return rv;
						InColorSpan = FALSE;
					}

					if ( CellColor != Args->DefaultColor ) {
						wsprintf( temp, L"<span class=\"X%02x\">", CellColor );
						if ( rv = BufferedOutput( OutBuffer, temp ) ) return rv;
						InColorSpan = TRUE;
					}

					LastColor = CellColor;
				}
			}

			if ( Entity )
				rv = BufferedOutput( OutBuffer, Entity );
			else
				rv = BufferedOutput( OutBuffer, ch );

			if ( rv )
				return rv;
		}

		if ( rv = BufferedOutput( OutBuffer, Args->Eol ) ) return rv;
	}

	if ( InColorSpan ) {
		if ( rv = BufferedOutput( OutBuffer, L"</span>" ) ) return rv;
		InColorSpan = FALSE;
		LastColor = Args->DefaultColor;
	}

	if ( rv = SaveCon_HtmlEndBody( OutBuffer, Args->Eol ) ) return rv;

	return FlushOutBuffer( OutBuffer );
}



#ifdef SAVECON_RTF_SUPPORT


int SaveCon_RtfCharEncode( wchar_t ch, LPTSTR outRep )
{
	if ( !ch || !outRep )
		return -1;


	if ( ( ch == '\\' ) || ( ch == '{' ) || ( ch == '}' ) ) {
		outRep[0] = '\\';
		outRep[1] = ch;
		outRep[2] = '\0';
	}
	else if ( ch == 0xa0 )
		wcscpy_s( outRep, SHORT_BUF_LEN, L"\\~" );
	else if ( ch < 0x0080 )
		outRep[0] = '\0';
	else if ( ch < 0x8000 )
		wsprintf( outRep, L"\\u%u ", ch );
	else
		wsprintf( outRep, L"\\u-%u ", 0x10000 - ch );

	return 0;
}



int SaveCon_RtfEncodeString( LPTSTR inString, LPTSTR outString, size_t buflen )
{
	if ( !inString || !outString )
		return -1;


	wchar_t temp[MAX_ARG_LEN]			= L"";
	wchar_t CharRep[SHORT_BUF_LEN]		= L"";
	size_t j							= 0;


	for ( size_t i = 0; inString[i]; i++ ) {
		SaveCon_RtfCharEncode( inString[i], CharRep );
		
		if ( CharRep[0] ) {
			wcscat_s( temp, MAX_ARG_LEN, CharRep );
			j = wcslen( temp );
		}
		else {
			if ( j < ( MAX_ARG_LEN - 1 ) ) {
				temp[j++] = inString[i];
				temp[j] = '\0';
			}
		}
	}

	if ( wcslen( temp ) >= buflen )
		return ERROR_NOT_ENOUGH_MEMORY;

	wcscpy_s( outString, buflen, temp );
	return 0;
}



int SaveCon_RtfEscapify( LPTSTR inString, LPTSTR outString, size_t buflen )
{
	if ( !inString || !outString )
		return -1;


	wchar_t temp[MAX_ARG_LEN]		= L"";
	int j							= 0;


	for ( size_t i = 0; inString[i]; i++ ) {
		if ( ( inString[i] == '\\' ) || ( inString[i] == '{' ) || ( inString[i] == '}' ) ) {
			if ( j >= ( MAX_ARG_LEN - 2 ) )
				return -2;

			temp[j++] = '\\';
		}

		if ( j >= ( MAX_ARG_LEN - 1 ) )
			return -2;

		temp[j++] = inString[i];
	}

	temp[j] = '\0';

	if ( j >= (int) buflen )
		return -2;

	wcscpy_s( outString, buflen, temp );
	return 0;
}



int SaveCon_SmashToAscii( LPTSTR inString, LPTSTR outString, size_t buflen )
{
	if ( !inString || !outString )
		return -1;


	wchar_t temp[MAX_ARG_LEN]		= L"";
	int j							= 0;
	BOOL found						= FALSE;


	outString[0]		= '\0';

	for ( size_t i = 0; inString[i]; i++ ) {

		if ( inString[i] < 0x0080 ) {
			if ( j < ( MAX_ARG_LEN - 1 ) ) {
				temp[j++] = inString[i];
				temp[j] = '\0';
			}
		}
		else {
			found = FALSE;

			for ( int k = 0; k < SAVECON_NUM_SMASH_CHARS; k++ ) {
				if ( inString[i] == SaveCon_SmashChars[k].ch ) {
					found = TRUE;

					for ( int q = 0; SaveCon_SmashChars[k].ascii[q]; q++ ) {
						if ( j < ( MAX_ARG_LEN - 1 ) ) {
							temp[j++] = SaveCon_SmashChars[k].ascii[q];
							temp[j] = '\0';
						}
					}
				}
			}

			if ( !found ) {
				if ( j < ( MAX_ARG_LEN - 1 ) ) {
					temp[j++] = '?';
					temp[j] = '\0';
				}
			}
		}
	}

	if ( j >= (int) buflen )
		return -2;


	wcscpy_s( outString, buflen, temp );
	return j;
}



int SaveCon_WriteRtfUnicodeDestination( BufOut_Buffer *OutBuffer, LPTSTR Keyword, LPTSTR Text )
{
	if ( !OutBuffer || !Keyword || !Text )
		return -1;


	wchar_t UnicodeText[MAX_ARG_LEN]		= L"";
	wchar_t AsciiText[MAX_ARG_LEN]			= L"";

	int k									= 0;
	int rv									= 0;


	for ( int i = 0; Text[i]; i++ )
		if ( Text[i] >= 0x80 )
			k++;

	if ( !k )
		goto AllAscii;


	SaveCon_RtfEncodeString( Text, UnicodeText, MAX_ARG_LEN );
	SaveCon_SmashToAscii( Text, AsciiText, MAX_ARG_LEN );
	SaveCon_RtfEscapify( AsciiText, AsciiText, MAX_ARG_LEN );


	if ( rv = BufferedOutput( OutBuffer, L"{\\upr{" ) )
		return rv;

	if ( rv = BufferedOutput( OutBuffer, Keyword ) )
		return rv;

	if ( rv = BufferedOutput( OutBuffer, AsciiText ) )
		return rv;

	if ( rv = BufferedOutput( OutBuffer, L"}{\\*\\ud{" ) )
		return rv;

	if ( rv = BufferedOutput( OutBuffer, Keyword ) )
		return rv;

	if ( rv = BufferedOutput( OutBuffer, UnicodeText ) )
		return rv;

	if ( rv = BufferedOutput( OutBuffer, L"}}}" ) )
		return rv;

	return 0;


AllAscii:

	SaveCon_RtfEscapify( Text, AsciiText, MAX_ARG_LEN );

	if ( rv = BufferedOutput( OutBuffer, L"{" ) )
		return rv;

	if ( rv = BufferedOutput( OutBuffer, Keyword ) )
		return rv;

	if ( rv = BufferedOutput( OutBuffer, AsciiText ) )
		return rv;

	if ( rv = BufferedOutput( OutBuffer, L"}" ) )
		return rv;

	return 0;
}



int SaveCon_CreateRtfColorTable( BufOut_Buffer *OutBuffer, DWORD *ConsolePalette, LPCTSTR Eol )
{
	wchar_t temp[SHORT_BUF_LEN]		= L"";
	DWORD color						= 0;
	int rv							= 0;


	wsprintf( temp, L"{\\colortbl%s;%s", Eol, Eol );
	if ( rv = BufferedOutput( OutBuffer, temp ) )
		return rv;

	for ( unsigned i = 0; i < 16; i++ ) {
		color = ConsolePalette[RtfToEgaColor[i]];
		wsprintf( temp, L"\\red%u\\green%u\\blue%u;%s", ( color & 0xff0000 ) >> 16, ( color & 0x00ff00 ) >> 8, color & 0x0000ff, Eol );

		if ( rv = BufferedOutput( OutBuffer, temp ) )
			return rv;
	}
	
	wsprintf( temp, L"}%s", Eol );
	return BufferedOutput( OutBuffer, temp );
}



int SaveCon_RtfHeader( BufOut_Buffer *OutBuffer, unsigned int Flags, LPTSTR Title, _SaveConArgs *Args )
{
	wchar_t temp[SHORT_BUF_LEN]		= L"";
	wchar_t verstr[SHORT_BUF_LEN]	= L"";
	wchar_t asciibuf[MAX_ARG_LEN]	= L"";

	ULONG buflen					= 0;
	int bg							= ( Args->DefaultColor & 0x00f0 ) >> 4;
	unsigned int FontSize			= Args->FontSize;

	int rv							= 0;



	if ( ( FontSize < SAVECON_FONT_SIZE_MIN ) || ( FontSize > SAVECON_FONT_SIZE_MAX ) )
		FontSize = SAVECON_FONT_SIZE_DEFAULT;


	if ( rv = BufferedOutput( OutBuffer, L"{\\rtf1\\ansi\\uc0\\deff0" ) )
		return rv;

	wsprintf( temp, L"\\deflang%u", 0x0400 );		//	GetUserDefaultUILanguage()
	if ( rv = BufferedOutput( OutBuffer, temp ) )
		return rv;

	wsprintf( temp, L"%s{\\fonttbl%s{\\f0\\fmodern\\fprq1 Consolas;{\\*\\falt Courier New;}}%s}%s", Args->Eol, Args->Eol, Args->Eol, Args->Eol );
	if ( rv = BufferedOutput( OutBuffer, temp ) )
		return rv;


	if ( Args->AttrBuffer ) {
		if ( rv = SaveCon_CreateRtfColorTable( OutBuffer, Args->Palette, Args->Eol ) )
			return rv;
		
		wsprintf( temp, L"{\\*\\background {\\shp{\\*\\shpinst\\shpleft0\\shptop0\\shpright0\\shpbottom0\\shpbxmargin\\shpbymargin\\shpwr0\\shpblwtxt1\\shpz0%s", Args->Eol );
		if ( rv = BufferedOutput( OutBuffer, temp ) )
			return rv;

		wsprintf( temp, L"{\\sp{\\sn shapeType}{\\sv 1}}{\\sp{\\sn fillColor}{\\sv %u}}{\\sp{\\sn fLine}{\\sv 0}}{\\sp{\\sn fBackground}{\\sv 1}}", SaveCon_Rgb2Bgr( Args->Palette[bg] ) );
		if ( rv = BufferedOutput( OutBuffer, temp ) )
			return rv;

		wsprintf( temp, L"}}}%s\\viewbksp1%s", Args->Eol, Args->Eol );
		if ( rv = BufferedOutput( OutBuffer, temp ) )
			return rv;
	}

	if ( ( Flags & SAVECON_FLAG_NO_METADATA ) == 0 ) {

		//		Open the info block :

		if ( rv = BufferedOutput( OutBuffer, L"{\\info" ) )
			return rv;

		//		Title :

		if ( rv = SaveCon_WriteRtfUnicodeDestination( OutBuffer, L"\\title ", Title ) )
			return rv;

		//		Author name :

		temp[0] = '\0';
		buflen = SHORT_BUF_LEN,
		GetUserNameExW( NameDisplay, temp, &buflen );

		if ( temp[0] == '\0' ) {
			buflen = SHORT_BUF_LEN;
			GetUserNameExW( NameUserPrincipal, temp, &buflen );
		}

		if ( temp[0] )
			if ( rv = SaveCon_WriteRtfUnicodeDestination( OutBuffer, L"\\author ", temp ) )
				return rv;

		//		Comment (software version) :

		MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, MACRO_STR( SAVECONSOLE_VERSION ), -1, verstr, SHORT_BUF_LEN );
		wsprintf( temp, L"{\\doccomm SaveConsole v%s (%s)}", verstr, PLATFORM );
		if ( rv = BufferedOutput( OutBuffer, temp ) )
			return rv;

		//		Creation time :

		wsprintf( temp, L"{\\creatim\\yr%u\\mo%u\\dy%u\\hr%u\\min%u}", Args->Now.wYear, Args->Now.wMonth, Args->Now.wDay, Args->Now.wHour, Args->Now.wMinute );
		if ( rv = BufferedOutput( OutBuffer, temp ) )
			return rv;

		//		Close the info block :

		if ( rv = BufferedOutput( OutBuffer, '}' ) )
			return rv;

		if ( rv = BufferedOutput( OutBuffer, Args->Eol ) )
			return rv;
	}

	wsprintf( temp, L"\\fs%u%s", FontSize * 2, Args->Eol );
	if ( rv = BufferedOutput( OutBuffer, temp ) )
		return rv;

	return 0;
}



int SaveCon_RtfEndBody( BufOut_Buffer *OutBuffer, LPCTSTR Eol )
{
	int rv				= 0;

	if ( rv = BufferedOutput( OutBuffer, '}' ) )
		return rv;

	return BufferedOutput( OutBuffer, Eol );
}



int SaveCon_RtfChangeColor( BufOut_Buffer *OutBuffer, WORD *OldColor, WORD NewColor )
{
	wchar_t temp[SHORT_BUF_LEN]			= L"";

	int oldbg							= -1;
	int oldfg							= -1;
	int bg								= ( NewColor & 0x00f0 ) >> 4;
	int fg								= NewColor & 0x000f;
	int rv								= 0;


	NewColor &= 0x00ff;

	if ( *OldColor != 0xffff ) {
		oldbg	= ( *OldColor & 0x00f0 ) >> 4;
		oldfg	= *OldColor & 0x000f;
	}

	if ( bg != oldbg ) {
		wsprintf( temp, L"\\highlight%u", EgaToRtfColor[bg] );
		if ( rv = BufferedOutput( OutBuffer, temp ) )
			return rv;
	}

	if ( fg != oldfg ) {
		wsprintf( temp, L"\\cf%u", EgaToRtfColor[fg] );
		if ( rv = BufferedOutput( OutBuffer, temp ) )
			return rv;
	}

	if ( NewColor != *OldColor ) {
		*OldColor = NewColor;
		return BufferedOutput( OutBuffer, ' ' );
	}

	return 0;
}



int SaveCon_Rtf( HANDLE ConsoleHandle, BufOut_Buffer *OutBuffer, unsigned int Flags, _SaveConArgs *Args )
{
	wchar_t Title[MAX_ARG_LEN]			= L"";
	wchar_t RtfTitle[MAX_ARG_LEN]		= L"";
	wchar_t temp[SHORT_BUF_LEN]			= L"";
	wchar_t CharRep[SHORT_BUF_LEN]		= L"";

	BOOL SaveColors						= Args->AttrBuffer ? TRUE : FALSE;

	unsigned int NumColumns				= Args->Columns;

	wchar_t ch							= '\0';
	WORD CellColor						= 0;
	WORD LastColor						= -1;

	int rv								= 0;



	SaveCon_MakeTitle( Title, &Args->Now );
	SaveCon_RtfEncodeString( Title, RtfTitle, MAX_ARG_LEN );


	if ( rv = SaveCon_RtfHeader( OutBuffer, Flags, Title, Args ) )
		return rv;


	if ( Flags & SAVECON_FLAG_TITLE ) {

		if ( rv = SaveCon_RtfChangeColor( OutBuffer, &LastColor, Args->DefaultColor ) )
			return rv;

		if ( rv = BufferedOutput( OutBuffer, L"\\b " ) )
			return rv;

		if ( rv = BufferedOutput( OutBuffer, RtfTitle ) )
			return rv;

		if ( rv = BufferedOutput( OutBuffer, L"\\b0\\line\\line" ) )
			return rv;

		if ( rv = BufferedOutput( OutBuffer, Args->Eol ) )
			return rv;
	}


	for ( unsigned int i = 0; i < Args->Rows; i++ ) {
		NumColumns = Args->Columns;

		while ( NumColumns && ( Args->CharBuffer[( i * Args->Columns ) + NumColumns - 1] == SPACE ) )
			NumColumns--;

		if ( Flags & SAVECON_FLAG_LINE_NUMBERS ) {
			if ( SaveColors && ( LastColor != Args->DefaultColor ) )
				if ( rv = SaveCon_RtfChangeColor( OutBuffer, &LastColor, Args->DefaultColor ) )
					return rv;

			wsprintf( temp, L"%4u:  ", i + 1 );
			if ( rv = BufferedOutput( OutBuffer, temp ) )
				return rv;
		}

		for ( unsigned int j = 0; j < NumColumns; j++ ) {
			ch = Args->CharBuffer[( i * Args->Columns ) + j];

			if ( SaveColors ) {
				CellColor = Args->AttrBuffer[( i * Args->Columns ) + j] & 0x00ff;

				if ( CellColor != LastColor )
					if ( rv = SaveCon_RtfChangeColor( OutBuffer, &LastColor, CellColor ) )
						return rv;
			}

			SaveCon_RtfCharEncode( ch, CharRep );

			if ( CharRep[0] )
				rv = BufferedOutput( OutBuffer, CharRep );
			else
				rv = BufferedOutput( OutBuffer, ch );

			if ( rv )
				return rv;
		}
		
		if ( rv = BufferedOutput( OutBuffer, L"\\line" ) )
			return rv;

		if ( rv = BufferedOutput( OutBuffer, Args->Eol ) )
			return rv;
	}

	if ( rv = SaveCon_RtfEndBody( OutBuffer, Args->Eol ) )
		return rv;

	return FlushOutBuffer( OutBuffer );
}


#endif	//	SAVECON_RTF_SUPPORT



int SaveCon_Text( HANDLE ConsoleHandle, BufOut_Buffer *OutBuffer, unsigned int Flags, _SaveConArgs *Args )
{
	wchar_t Title[MAX_ARG_LEN]		= L"";
	wchar_t temp[SHORT_BUF_LEN]		= L"";
	wchar_t LongChar[3]				= L"";

	unsigned int NumColumns			= Args->Columns;
	wchar_t ch						= '\0';

	int rv							= 0;


	if ( Flags & SAVECON_FLAG_TITLE ) {
		SaveCon_MakeTitle( Title, &Args->Now );
		wcscat_s( Title, MAX_ARG_LEN, Args->Eol );
		wcscat_s( Title, MAX_ARG_LEN, Args->Eol );
	}


	if ( ( Flags & SAVECON_FLAG_NO_BOM ) == 0 )
		BufferedOutput( OutBuffer, BOM );

	if ( Flags & SAVECON_FLAG_TITLE )
		if ( rv = BufferedOutput( OutBuffer, Title ) ) return rv;


	for ( unsigned int i = 0; i < Args->Rows; i++ ) {
		NumColumns = Args->Columns;
		if ( ( Flags & SAVECON_FLAG_FULL_WIDTH ) == 0 )
			while ( NumColumns && ( Args->CharBuffer[( i * Args->Columns ) + NumColumns - 1] == SPACE ) )
				NumColumns--;

		if ( Flags & SAVECON_FLAG_LINE_NUMBERS ) {
			wsprintf( temp, L"%4u:  ", i + 1 );
			if ( rv = BufferedOutput( OutBuffer, temp ) ) return rv;
		}

		for ( unsigned int j = 0; j < NumColumns; j++ ) {
			ch = Args->CharBuffer[( i * Args->Columns ) + j];

			//	Make sure we're not going to split a surrogate pair,
			//	as that would bollix UTF-16 to UTF-8 conversion : 

			if ( ( ( ch & 0xfc00 ) == 0xd800 ) && ( j < ( NumColumns - 1 ) ) )
				if ( ( Args->CharBuffer[( i * Args->Columns ) + j + 1] & 0xfc00 ) == 0xdc00 )
					if ( OutBuffer->ptr > ( OutBuffer->size - 6 ) )
						if ( rv = FlushOutBuffer( OutBuffer ) ) return rv;

			if ( rv = BufferedOutput( OutBuffer, ch ) ) return rv;
		}

		if ( rv = BufferedOutput( OutBuffer, Args->Eol ) ) return rv;
	}

	return FlushOutBuffer( OutBuffer );
}



DLLExports int WINAPI saveconsole( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t Filename[MAX_FILENAME_LEN]		= L"";
	wchar_t Ext[SHORT_BUF_LEN]				= L"";
	wchar_t ErrorText[SHORT_BUF_LEN]		= L"";
	const wchar_t ModeStr[3][5]				= { L"text", L"HTML", L"RTF" };

	int OutputEncoding					= BUFOUT_FILE_UTF8;
	unsigned int Flags					= 0;
	unsigned int Mode					= SAVECON_MODE_DEFAULT;
	BOOL NoColors						= FALSE;
	BOOL AutoOpen						= FALSE;
	BOOL Quiet							= FALSE;
	BOOL ViewportOnly					= FALSE;
	BOOL UseConsolePalette				= FALSE;
	BOOL ToClipboard					= FALSE;
	BOOL NoDialog						= FALSE;
	BOOL UseHook						= TRUE;
	unsigned HookProcFlags				= 0;

	HANDLE ConsoleHandle				= NULL;
	CONSOLE_SCREEN_BUFFER_INFO ConsoleInfo;
	DWORD ConsolePalette[16];

	BufOut_Buffer *OutBuffer			= NULL;

	const int LinesPerChunk				= 32;
	unsigned int line2					= 0;
	COORD ReadCoord;
	DWORD CellsRead						= 0;

	unsigned int FirstRow				= 0;
	unsigned int LastRow				= 0;

	unsigned __int64 StartTime			= GetTickCount2();
	unsigned __int64 Duration			= 0;
	unsigned __int64 FileSize			= 0;
	unsigned __int64 TotalChars			= 0;
	unsigned __int64 TotalLines			= 0;
	unsigned int NumWrites				= 0;

#ifdef SAVECON_INTVARS
	LPCTSTR LastFile					= NULL;
#endif	//	SAVECON_INTVARS

	_SaveConArgs Args;

	int rv								= 1;



	memset( &Args, 0x00, sizeof( _SaveConArgs ) );

	Args.DefaultColor	= SAVECON_DEFAULT_COLOR;
	Args.BackColor		= SAVECON_DEFAULT_BG_COLOR;
	Args.Eol[0]			= '\n';
	Args.Palette		= (DWORD*) CssColors;
	GetLocalTime( &Args.Now );


#ifdef SAVECON_REGISTRY_DEFAULTS

	SaveCon_GetDefaults( &Flags, &OutputEncoding, Args.Eol, &NoDialog );

#endif	//	SAVECON_REGISTRY_DEFAULTS


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {

		if ( !wcsncmp( arg[i], L"/?", 2 ) )
			return ShowCmdHelp( SaveConsoleHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( SAVECONSOLE_VERSION ) );
		else if ( !_wcsnicmp( arg[i], L"/BG", 3 ) ) {
			int j = 3;
			if ( IsColonOrEquals( arg[i][j] ) )
				j++;

			Args.BackColor = W3CColorValue( arg[i] + j, TRUE );
			if ( Args.BackColor == W3C_COLOR_UNDEFINED ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad color name: \"%s\"\r\n", arg[i] + j );
				goto CleanupAndExit;
			}
		}
		else if ( !_wcsicmp( arg[i], L"/CP" ) )
			UseConsolePalette = TRUE;
		else if ( !_wcsicmp( arg[i], L"/CR" ) )
			Args.Eol[0] = '\r', Args.Eol[1] = '\0';
		else if ( !_wcsicmp( arg[i], L"/CRLF" ) )
			Args.Eol[0] = '\r', Args.Eol[1] = '\n', Args.Eol[2] = '\0';
		else if ( !_wcsicmp( arg[i], L"/D" ) )
			NoColors = TRUE;
		else if ( !_wcsnicmp( arg[i], L"/FS", 3 ) ) {
			int j = 3;
			if ( IsColonOrEquals( arg[i][j] ) )
				j++;

			if ( ParseInt( arg[i] + j, &Args.FontSize, NULL ) ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad font size: \"%s\"\r\n", arg[i] + j );
				goto CleanupAndExit;
			}

			if ( ( Args.FontSize < SAVECON_FONT_SIZE_MIN ) || ( Args.FontSize > SAVECON_FONT_SIZE_MAX ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Font size must be %u to %u: %u\r\n", SAVECON_FONT_SIZE_MIN, SAVECON_FONT_SIZE_MAX, Args.FontSize );
				goto CleanupAndExit;
			}
		}
		else if ( !_wcsicmp( arg[i], L"/H" ) )
			Mode = SAVECON_MODE_HTML;
		else if ( !_wcsicmp( arg[i], L"/L" ) )
			Flags |= SAVECON_FLAG_LINE_NUMBERS;
		else if ( !_wcsicmp( arg[i], L"/LF" ) )
			Args.Eol[0] = '\n', Args.Eol[1] = '\0';
		else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
			int j = 2;
			if ( IsColonOrEquals( arg[i][j] ) )
				j++;

			while ( arg[i][j] ) {
				switch ( towupper( arg[i][j] ) ) {
				case 'B' :
					Flags |= SAVECON_FLAG_NO_BOM;
					break;

				case 'C' :
					NoColors = TRUE;
					break;

				case 'D' :
					NoDialog = TRUE;
					break;

				case 'F' :
					Quiet = TRUE;
					break;

				case 'H' :
					UseHook = FALSE;
					break;

				case 'I' :
					Flags |= SAVECON_FLAG_NO_METADATA;
					break;

				case 'T' :
					HookProcFlags |= SAVECON_HP_FLAG_DISABLE_TYPE;
					break;

				case 'Y' :
					HookProcFlags |= SAVECON_HP_FLAG_DISABLE_TANYA;
					break;

				case '?' :
					QPuts( (LPTSTR) SaveConsoleOptNHelpText );
					return 0;

				default :
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Unknown suboption: /N:%c\r\n", towupper( arg[i][j] ) );
					goto CleanupAndExit;
				}

				j++;
			}
		}
		else if ( !_wcsicmp( arg[i], L"/O" ) )
			AutoOpen = TRUE;
		else if ( !_wcsicmp( arg[i], L"/Q" ) )
			Quiet = TRUE;

#ifdef SAVECON_RTF_SUPPORT

		else if ( !_wcsicmp( arg[i], L"/R" ) )
			Mode = SAVECON_MODE_RTF;

#endif	//	SAVECON_RTF_SUPPORT

		else if ( !_wcsicmp( arg[i], L"/T" ) )
			Flags |= SAVECON_FLAG_TITLE;
		else if ( !_wcsnicmp( arg[i], L"/U", 2 ) ) {
			int j = 2;

			if ( !_wcsnicmp( arg[i] + j, L"TF", 2 ) )
				j += 2;

			if ( arg[i][j] == '-' )
				j++;

			if ( !_wcsicmp( arg[i] + j, L"8" ) )
				OutputEncoding = BUFOUT_FILE_UTF8;
			else if ( !_wcsicmp( arg[i] + j, L"16" ) )
				OutputEncoding = BUFOUT_FILE_UTF16;
			else {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				goto CleanupAndExit;
			}
		}
		else if ( !_wcsicmp( arg[i], L"/V" ) )
			ViewportOnly = TRUE;
		else if ( !_wcsicmp( arg[i], L"/W" ) )
			Flags |= SAVECON_FLAG_FULL_WIDTH;
		else if ( !_wcsnicmp( arg[i], L"/X", 2 ) ) {
			int j = 2;
			if ( IsColonOrEquals( arg[i][j] ) )
				j++;

			if ( arg[i][j] )
				Args.StyleSheet = arg[i] + j;
			else
				Args.StyleSheet = NULL;
		}
		else if ( !_wcsnicmp( arg[i], L"/Y", 2 ) ) {
			int j = 2;
			if ( IsColonOrEquals( arg[i][j] ) )
				j++;

			if ( arg[i][j] )
				Args.EmbedStyle = arg[i] + j;
			else
				Args.EmbedStyle = NULL;
		}
		else if ( arg[i][0] == '/' ) {
			_wcsupr_s( arg[i], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
			goto CleanupAndExit;
		}
		else if ( Filename[0] == '\0' ) {
			if ( !_wcsicmp( arg[i], L"clip:" ) ) {
				SaveCon_MakeFilename( Filename );
				ToClipboard = TRUE;
			}
			else {

#ifdef		FILE_HANDLER_VERSION
				FH_Truename( arg[i], Filename );
#else	//	FILE_HANDLER_VERSION
				QueryTrueName( arg[i], Filename );
#endif	//	FILE_HANDLER_VERSION

				if ( Filename[0] == '\0' ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad filename: \"%s\"\r\n", arg[i] );
					goto CleanupAndExit;
				}
			}
		}
		else {
			_wcsupr_s( arg[i], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
			goto CleanupAndExit;
		}
	}


	SaveCon_ExtensionPart( Filename, Ext );

	if ( !Mode ) {
		if ( !_wcsicmp( Ext, L".html" ) || !_wcsicmp( Ext, L".htm" ) )
			Mode = SAVECON_MODE_HTML;

#ifdef SAVECON_RTF_SUPPORT

		else if ( !_wcsicmp( Ext, L".rtf" ) )
			Mode = SAVECON_MODE_RTF;

#endif	//	SAVECON_RTF_SUPPORT

	}


	if ( !NoDialog && ( ( Filename[0] == '\0' ) || QueryIsDirectory( Filename ) ) ) {
		if ( SaveCon_GetFilename( Filename, MAX_FILENAME_LEN, UseHook,&Mode, HookProcFlags ) == 3 ) {
			GetErrorText( ERROR_CANCELLED, ErrorText );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"%s\r\n", ErrorText );
			rv = 3;
			goto CleanupAndExit;
		}
	}


	if ( Filename[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename\r\n" );
		goto CleanupAndExit;
	}


#ifdef SAVECON_RTF_SUPPORT

	if ( Mode == SAVECON_MODE_RTF ) {
		OutputEncoding = BUFOUT_FILE_OEM;
		Args.Palette = (DWORD*) RtfColors;
	}

#endif	//	SAVECON_RTF_SUPPORT


	if ( ToClipboard ) {
		OutputEncoding = BUFOUT_FILE_UTF16;
		Flags |= SAVECON_FLAG_NO_BOM;
		Args.Eol[0]	= '\r';
		Args.Eol[1]	= '\n';
		Args.Eol[2] = '\0';
		AutoOpen = FALSE;
	}

	
	rv = 2;


	//	Open a handle to the console :

	ConsoleHandle = CreateFile( L"CONOUT$", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL );

	if ( ConsoleHandle == INVALID_HANDLE_VALUE ) {
		GetLastErrorText( ErrorText );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error opening console:\r\n" );
		Qprintf( ERRHANDLE, L"    %s\r\n", ErrorText );
		goto CleanupAndExit;
	}


	//	Get buffer info :

	memset( &ConsoleInfo, 0x00, sizeof( CONSOLE_SCREEN_BUFFER_INFO ) );

	if ( !GetConsoleScreenBufferInfo( ConsoleHandle, &ConsoleInfo ) ) {
		GetLastErrorText( ErrorText );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error getting console info:\r\n" );
		Qprintf( ERRHANDLE, L"    %s\r\n", ErrorText );
		goto CleanupAndExit;
	}

	if ( ( ConsoleInfo.dwSize.X < 10 ) || ( ConsoleInfo.dwSize.Y < 5 ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Invalid console size!\r\n" );
		goto CleanupAndExit;
	}


	//	Set coordinate and size variables :

	if ( ViewportOnly ) {
		FirstRow		= ConsoleInfo.srWindow.Top;
		LastRow			= ConsoleInfo.srWindow.Bottom;
	}
	else {
		FirstRow		= 0;
		LastRow			= ConsoleInfo.dwCursorPosition.Y;

		if ( (unsigned) ConsoleInfo.srWindow.Bottom > LastRow )
			LastRow = ConsoleInfo.srWindow.Bottom;
	}

	Args.Columns		= ConsoleInfo.dwSize.X;
	Args.Rows			= LastRow - FirstRow + 1;

	if ( !NoColors )
		Args.DefaultColor	= ConsoleInfo.wAttributes & 0x00ff;


	//	Allocate buffers for characters and attributes :

	Args.CharBuffer = (wchar_t*) malloc( Args.Rows * Args.Columns * sizeof( wchar_t ) );
	if ( Args.CharBuffer == NULL ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can\u2019t allocate characters buffer\r\n" );
		goto CleanupAndExit;
	}

	if ( Mode && !NoColors ) {
		Args.AttrBuffer = (WORD*) malloc( Args.Rows * Args.Columns * sizeof( WORD ) );
		if ( Args.AttrBuffer == NULL ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Can\u2019t allocate attributes buffer\r\n" );
			goto CleanupAndExit;
		}
	}


	//	Slurp the console data into the buffers :

	for ( unsigned int line = FirstRow; line <= LastRow; line += LinesPerChunk ) {
		line2 = line + LinesPerChunk -1;
		if ( line2 > LastRow )
			line2 = LastRow;

		ReadCoord.X = 0;
		ReadCoord.Y = line;

		if ( !ReadConsoleOutputCharacter( ConsoleHandle, Args.CharBuffer + ( ( line - FirstRow ) * Args.Columns ), ( line2 - line + 1 ) * Args.Columns, ReadCoord, &CellsRead ) ) {
			GetLastErrorText( ErrorText );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error reading characters: %s\r\n", ErrorText );
			goto CleanupAndExit;
		}

		if ( Args.AttrBuffer ) {
			if ( !ReadConsoleOutputAttribute( ConsoleHandle, Args.AttrBuffer + ( ( line - FirstRow ) * Args.Columns ), ( line2 - line + 1 ) * Args.Columns, ReadCoord, &CellsRead ) ) {
				GetLastErrorText( ErrorText );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Error reading attributes: %s\r\n", ErrorText );
				goto CleanupAndExit;
			}

			if ( UseConsolePalette )
				if ( SaveCon_GetConsolePalette( ConsoleHandle, ConsolePalette ) )
					Args.Palette = ConsolePalette;
		}
	}


	//	Open a buffered output file :

	OutBuffer = OpenBufferedFile( Filename, SAVECON_BUFFER_SIZE, OutputEncoding, 0 );
	if ( !OutBuffer ) {
		GetErrorText( GetLastError(), ErrorText );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error opening output file: \"%s\"\r\n", Filename );
		Qprintf( ERRHANDLE, L"%s\r\n", ErrorText );
		goto CleanupAndExit;
	}

	if ( IsConsoleHandle( OutBuffer->handle ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Console filename not allowed: \"%s\"\r\n", Filename );
		goto CleanupAndExit;
	}


	//	Save the console data to the output file :

	switch ( Mode ) {

	case SAVECON_MODE_HTML :
		rv = SaveCon_Html( ConsoleHandle, OutBuffer, Flags, &Args );
		break;

#ifdef SAVECON_RTF_SUPPORT

	case SAVECON_MODE_RTF :
		rv = SaveCon_Rtf( ConsoleHandle, OutBuffer, Flags, &Args );
		break;

#endif	//	SAVECON_RTF_SUPPORT

	default :
		rv = SaveCon_Text( ConsoleHandle, OutBuffer, Flags, &Args );
	}

	if ( rv ) {
		GetErrorText( rv, ErrorText );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error saving to \"%s\":\r\n", Filename );
		Qprintf( ERRHANDLE, L"   %s\r\n", ErrorText );
		rv = 2;
		goto CleanupAndExit;
	}

	
	//	Save info from the buffered-file structure, and close the output file :

	TotalLines	= OutBuffer->lines;
	TotalChars	= OutBuffer->chars;
	FileSize	= OutBuffer->bytes;
	NumWrites	= OutBuffer->writes;
	CloseBufferedFile( OutBuffer );
	OutBuffer	= NULL;


	//	If it's going to the clipboard, handle that, then delete the temp file :

	if ( ToClipboard ) {

		if ( !SaveCon_CopyToClipboard( Filename, CF_UNICODETEXT ) ) {
			GetLastErrorText( ErrorText );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error copying \"%s\" to CLIP:\r\n", Filename );
			Qprintf( ERRHANDLE, L"   %s\r\n", ErrorText );
			DeleteFile( Filename );
			rv = 2;
			goto CleanupAndExit;
		}
		
		DeleteFile( Filename );
	}


	//	Launch the output file, if desired :

	if ( AutoOpen )
		SaveCon_LaunchFile( Filename );


	//	Display the final report :

	if ( !Quiet ) {
		wchar_t rowsbuf[64]		= L"";
		wchar_t colsbuf[64]		= L"";
		wchar_t timebuf[64]		= L"";
		wchar_t sizebuf[64]		= L"";
		wchar_t linesbuf[64]	= L"";
		wchar_t charsbuf[64]	= L"";
		wchar_t writesbuf[64]	= L"";

		Duration = GetTickCount2() - StartTime;
		wsprintf( timebuf, L"%I64u.%03u", Duration / 1000, Duration % 1000 );

		SaveCon_Commafy( Args.Rows, rowsbuf );
		SaveCon_Commafy( Args.Columns, colsbuf );
		SaveCon_Commafy( FileSize, sizebuf );
		SaveCon_Commafy( TotalLines, linesbuf );
		SaveCon_Commafy( TotalChars, charsbuf );
		SaveCon_Commafy( NumWrites, writesbuf );

		if ( ToClipboard )
			Printf( L"\r\nSaved to CLIP: as %s.\r\n", ModeStr[Mode] );
		else
			Printf( L"\r\nSaved to \"%s\" as %s (%s, %s).\r\n", Filename, ModeStr[Mode], ( OutputEncoding == BUFOUT_FILE_OEM ) ? L"ASCII" : BufOut_EncodingName( OutputEncoding ), SaveCon_EolType( Args.Eol ) );

		Printf( L"    Console area:  %s lines \u00d7 %s columns\r\n", rowsbuf, colsbuf );
		Printf( L"    Output file :  %s lines   %s characters   %s bytes\r\n", linesbuf, charsbuf, sizebuf );
		Printf( L"    Performance :  %s disk write%s    %s seconds\r\n", writesbuf, ( NumWrites == 1 ) ? L"" : L"s", timebuf );
	}


CleanupAndExit:

	free( Args.CharBuffer );
	free( Args.AttrBuffer );
	CloseHandle( ConsoleHandle );
	CloseBufferedFile( OutBuffer );
	
#ifdef SAVECON_INTVARS

	if ( rv == 0 ) {	//	the save succeeded:  return either the filename or clip:
		LastFile = ToClipboard ? SaveCon_ClipName: Filename;
		SC_LastMode = Mode;
	}
	else if ( ( rv == 1 ) || ( rv == 3 ) )		//	the save never started:  do not change the internal variables
		LastFile = NULL;
	else {				//	error on save:  return the error string
		LastFile = SaveCon_ErrorString;
		SC_LastMode = 999;
	}

	if ( LastFile ) {	//	update the internal variables :
		free( SC_LastFile );

		SC_LastFile = (LPTSTR) malloc( sizeof( wchar_t ) * ( wcslen( LastFile ) + 1 ) );
		if ( SC_LastFile )
			wcscpy_s( SC_LastFile, wcslen( LastFile ) + 1, LastFile );
	}

#endif	//	SAVECON_INTVARS

	return rv;
}



#ifdef SAVECON_INTVARS

DLLExports int _saveconfile( LPTSTR outString )
{
	if ( !outString )
		return 1;


	if ( SC_LastFile )
		wcscpy_s( outString, MAX_FILENAME_LEN, SC_LastFile );
	else
		outString[0] = '\0';

	return 0;
}



DLLExports int _saveconmode( LPTSTR outString )
{
	if ( !outString )
		return 1;


	const wchar_t ModeStr[3][5]				= { L"text", L"HTML", L"RTF" };

	if ( SC_LastMode == 999 )
		wcscpy_s( outString, MAX_ARG_LEN, SaveCon_ErrorString );
	else if ( SC_LastMode <= SAVECON_MODE_RTF )
		wcscpy_s( outString, MAX_ARG_LEN, ModeStr[SC_LastMode] );
	else
		outString[0] = '\0';

	return 0;
}

#endif	//	SAVECON_INTVARS

#endif	//	SAVECONSOLE_VERSION


/*

	1.4.0.0		2020-10-04
				Adds experimental RTF support.

	1.4.0.1		2020-10-17
				Tweaks ASCIIfication of the Euro symbol.

	1.4.0.2		2020-10-20
				Removes an undocumented minor feature (local language declaration in HTML
				and RTF).  Tweaks the syntax message.

	1.4.0.3		2020-10-22
				Implements internal replacement for CopyToClipboard(), which was failing
				on RTF files (misinterpreting UTF-16 as ANSI??)  Fixed erroneous
				encoding value in final report when saving RTF to the clipboard.

	1.4.1.0		2020-10-31
				Adds a 'Save As' dialog to be shown when the filename is not specified,
				or if it names a directory.

	1.4.1.1		2020-11-07
				Adds a count of characters written to the final report.  No BOM when
				saving to the clipboard.

	1.4.2.0		2020-11-16
				Adds /X: to insert a link to an external style sheet.  Changes the
				#define controlling the help text from SAVECON_FULL_HELP_TEXT to
				SAVECON_SHORT_HELP_TEXT.

	1.4.3.0		2020-12-01
				Adds /Y: to specify CSS for the PRE tag in the internal style sheet.
				Adds support for _SAVECONFILE and _SAVECONMODE variables to return the
				filename and mode of the last save.

	1.4.3.1		2021-05-04
				If you combine /FS: and /Y:, /Y: wins.

	1.4.3.2		2021-06-19
				A .HTM extension triggers /H mode, the same as .HTML.

	1.4.4.0		2021-10-16
				Adds /NT to disable the 'Save as type:' control.  Now changing the
				extension in the dialog's edit control can change the save format.
				Now _SAVECONMODE returns *ERROR* after a failed save, just like
				_SAVECONFILE.

	1.4.4.1		2021-10-31
				Help text is now displayed via ShowCmdHelp().  Tweaked the default
				background color a bit.

	1.4.4.2		2021-11-19
				Always null-terminate Args.Eol.  (Shouldn't be needed, but what the heck.)

	1.4.4.3		2021-11-24
				We don't need to entitize non-breaking spaces anymore, do we?

	1.4.4.4		2021-12-04
				A bug fix for the Clinton era:  Tweaked the CSS color for white; now
				it's web-safe.

	1.4.4.5		2021-12-21
				Fixed an issue calling TakeCommandIPC() on 32-bit TCC.

	1.4.4.6		2021-12-22
				No longer links to TakeCommandIPC() dynamically.  It's been a long time
				since that was necessary.

	1.4.4.7		2021-02-03
				Minor tweaks to my web-safe palette.  Bright blue is a bit brighter,
				dark yellow is a tad browner.

	1.4.5.0		2022-01-05
				Fix for trailing spaces with a different background color being lost.

	1.4.5.1		2022-06-14
				Made RTF support a compile-time option (#define SAVECON_RTF_SUPPORT
				to enable).

	1.4.5.2		2022-07-18
				Spelling correction.  No changes to functionality.

	1.4.5.3		2023-05-23
				Removed the trailing slash from the end of the META CHARSET tag.

	1.4.6.0		2023-06-13
				Removed calls to QueryIsFile() to work around an issue in TCC
				v30.00.20.

	1.4.7.0		2023-09-07
				Added support for registry defaults; the only values currently
				supported are "No BOM", "Output Encoding", "Line Ends", and
				"Disable SaveAs Dialog".
				Added SAVECONSOLE /N? help text.

	1.4.7.1		2023-09-08
				Disabled all the registry defaults stuff.  It's a pointless waste
				of space; aliases provide a perfectly good mechanism for providing
				default options.

*/
